;       Last change    August 1993
(in-package epilib)

;
; Complexity routines
;

(tweakable '*literal-complexity* 2  
  "Base complexity for each literal" '(float * *))
(tweakable '*variable-complexity* 2 
  "Complexity for a variable" '(float * *))
(tweakable '*constant-complexity* 5 ;1
  "Complexity for a constant" '(float * *))
(tweakable '*function-complexity* 2 
  "Complexity for a function or kind operator term" '(float * *))

(defun set-complexity (wff &aux c)
  (setq c (wff-complexity wff))
  (setp wff 'complexity c)
  )

; complexity calculation 
;   each literal has a base value of *literal-complexity*
;        variable                    *variable-complexity*
;        constant                    *constant-complexity*
;        function/kind-term          *function-complexity*
; for **, *, @, _ propositions, the complexity is the sum of the literal
;    complexity (first argument) and the episodic term (last argument)
; for and, or, implies, the complexity is the sum of the complexities of
;    the argument propositions
; for typical literals, the complexity is the base value, + the 
;    complexities of all arguments (variable, constant and functions)


(defun wff-complexity (wff &aux r)
  (case (getp wff 'wff-type)
	(constant *constant-complexity*)
	(variable *variable-complexity*)
	((logical episodic causal)
	 (apply '+ (mapcar 'item-complexity (getp wff 'arguments))))
	(quantified
	 (+ (if (setq r (getp wff 'restrict)) (wff-complexity r) 0)
	    (wff-complexity (getp wff 'main))))
	(modified 
	 (+ *function-complexity* 
	    (apply '+ (mapcar 'item-complexity (getp wff 'arguments)))))
	(prefix 
	 (+ *literal-complexity*
	    (pred-complexity (getp wff 'pred))
	    (apply '+ (mapcar 'item-complexity (getp wff 'arguments)))))
	(t *constant-complexity*))
  )

(defun item-complexity (item)
  (cond ((symbolp item)
	 (case (entity-type item)
	       (wff (wff-complexity item))
	       (term (term-complexity item))
	       (pred (pred-complexity item))
	       (op (op-complexity item))
	       (otherwise (term-complexity item))
	       ))
	(t *constant-complexity*))
  )

(defun term-complexity (term)
  (cond ((symbolp term)
	 (case (getp term 'term-type)
	       (constant *constant-complexity*)
	       (variable *variable-complexity*)
	       ((function modified wff) *function-complexity*)
	       ((record quoted-expression quasi-quoted-expression) 
		*constant-complexity*)
	       (otherwise *constant-complexity*)))
	(t *constant-complexity*))
  )

(defun pred-complexity (pred)
  (cond ((symbolp pred)
	 (case (getp pred 'pred-type)
	       ((constant variable) 0)
	       ((lambda modified) *function-complexity*)
	       (otherwise 0)))
	(t 0))
  )

(defun op-complexity (op)
  (cond ((symbolp op)
	 (case (getp op 'op-type)
	       ((constant variable) 0)
	       (modified *function-complexity*)
	       (otherwise 0)))
	(t 0))
  )

