;; 9 February 2005 - added a change-property call in create-predicate
;; to store the chages to the property SPECIALISTS of the
;; predicates. This allows a clean retraction in case of changes to
;; hierarchies. Thanks to Alexander Yeh to notice this problem. FM.

;;       Last change    August 1993
(in-package epilib)

; create-predicate adds a specialist to a predicate for use by the specialist
; interface
(defun create-predicate (predname specialist &aux
				  (s (getp predname 'SPECIALISTS)))
  (rewind-question)
  (change-property predname 'PREDICATE t)
  ;; added this change property to allow a clean retraction in case an
  ;; hierarchy is deleted - thanks to Alexander Yeh to notice this
  ;; problem. FM.
  (change-property predname 'SPECIALISTS (union s (list specialist)))
  (setp predname 'SPECIALISTS (union s (list specialist)))
  )

; create-function adds a specialist to a function for use by the specialist
; interface
; It also flags fnname as a function
(defun create-function (fnname specialist &aux
			       (s (getp fnname 'SPECIALISTS)))
  (rewind-question)
  (when specialist 
	(setp fnname 'SPECIALISTS (union s (list specialist))) )
  (if specialist
      (setp fnname 'FUNCTION t)
      (change-property fnname 'FUNCTION t))
  )

(defun create-operator (opname specialist &aux
			       (s (getp opname 'SPECIALISTS)))
  (rewind-question)
  (when specialist 
	(setp opname 'SPECIALISTS (union s (list specialist))) )
  (if specialist
      (setp opname 'OPERATOR t)
      (change-property opname 'OPERATOR t))
  )

(defun function-with-rel-pred (fn pred &optional sort)
  (rewind-question)
  (when pred (change-property fn 'REL-PRED pred))
  (when sort (change-property fn 'SORT sort))
  )

(defun other-function (package &rest fns)
  (rewind-question)
  ; adds functions to the "other" specialist
  (dolist (f fns)
          (create-function f 'other-specialist)
          (change-property f 'package package)
          )
  )

(defun other-predicate (package &rest preds)
  (rewind-question)
  (dolist (p preds)
          (create-predicate p 'other-specialist)
          (change-property p 'package package)
          )
  )

(defun other-entry-rtn (package pred entry-rtn)
  (rewind-question)
  (change-property pred 'entry-rtn (cond ((find-symbol (symbol-name entry-rtn) package)) 
					 (t entry-rtn)))
  )

