;       Last change    August 1993
(in-package epilib)

;;;; Routines for gathering information - useful for EPILOG and specialists

(tweakable '*quoted-indicators* '(quote qquote) 
  "list of symbols which indicate that what follows is a quoted list for EPILOG"
 'symbol-list )
(tweakable '*record-indicators* '($) 
  "list of symbols which indicate that what follows is record for EPILOG"
 'symbol-list )

(defvar *subs* nil)

; Information routines for use AFTER normalization
; These operate on formulas, terms, etc in their normalized atom form

(defmacro term-type (term) 
 `(if (symbolp ,term) (cond ((getp ,term 'term-type)) (t 'constant)) 'constant))
(defmacro pred-type (pred) 
 `(if (symbolp ,pred) (cond ((getp ,pred 'pred-type)) (t 'constant)) 'constant))
(defmacro op-type (op) 
 `(if (symbolp ,op) (cond ((getp ,op 'op-type)) (t 'constant)) 'constant))
(defmacro quant-type (q) 
 `(if (symbolp ,q) (cond ((getp ,q 'quant-type)) (t 'constant)) 'constant))
(defmacro wff-type (wff) 
 `(if (symbolp ,wff) (cond ((getp ,wff 'wff-type)) (t 'constant)) 'constant))
(defun entity-type (item) 
  (cond ((symbolp item)
	 (cond ((getp item 'entity-type))
	       ((equal (getp item 'operator) t) 'operator)
	       ((equal (getp item 'quantifier) t) 'quantifier)
	       ((or (getp item 'indicators) (getp item 'pred-type)
		    (getp item 'specialists)
		    (logical-conn item) (episodic-conn item) (causal-conn item))
		'pred)
	       (t 'term)))
	((numberp item) 'number)
	(t nil))
  )
(defmacro operator-result (item) `(if (symbolp ,item) (getp ,item 'result)))
(defmacro operator-args (item) `(if (symbolp ,item) (getp ,item 'args)))
(defmacro operator-type (item) 
  `(if (symbolp ,item) (getp ,item 'operator-type)))

(defmacro neg-p (wff) `(and (symbolp ,wff) (getp ,wff 'negated)))
(defmacro wff-p (wff) `(member (entity-type ,wff) '(wff)))
(defmacro pred-p (pred) `(member (entity-type ,pred) '(pred number)))
(defmacro term-p (term) `(member (entity-type ,term) '(term number)))
(defmacro operator-p (op) `(member (entity-type ,op) '(operator number)))
(defmacro quantifier-p (q) `(member (entity-type ,q) '(quantifier number)))
(defmacro function-p (f) `(and (symbolp ,f) (getp ,f 'function)))
(defmacro epi-variable-p (f) `(and (symbolp ,f) (equal (getp ,f 'variable) t)))
(defun epi-variable (f) (epi-variable-p f))
(defmacro function-term-p (f) 
  `(and (term-p ,f) (equal (term-type ,f) 'function)))
(defmacro lambda-pred-p (p) 
  `(and (pred-p ,p) (equal (pred-type ,p) 'lambda)))
(defmacro record-p (f) `(and (term-p ,f) (equal (term-type ,f) 'record)))
(defmacro quoted-expression-p (f) 
  `(and (term-p ,f) (equal (term-type ,f) 'quoted-expression)))
(defmacro quasi-quoted-expression-p (f) 
  `(and (term-p ,f) (equal (term-type ,f) 'quasi-quoted-expression)))
(defmacro type-pred-p (p) 
  `(and (symbolp ,p) (member 'type-specialist (getp ,p 'specialists))))
(defmacro skolem-constant-p (c) `(and (symbolp ,c) (getp ,c 'indef)))

(defun print-info (a) (cond ((symbolp a) (cond ((getp a 'print)) (t a))) (t a)))


; Information routines for use BEFORE and during normalization
; These operate on formulas, terms, etc in their list form

;; prop-pred returns the predicate of the proposition
;; if this is a quantified expression, no predicate is returned
(defun prop-pred (prop)
  (cond ((atom prop) nil)
        ((is-negated prop) (prop-pred (cdr prop)))
        (t (cadr prop)))
  )
 
;; prop-args returns a list of the arguments for the proposition
(defun prop-args (prop)                                          
  (cond ((atom prop) nil)
        ((is-negated prop) (prop-args (cdr prop)))
        (t (cons (car prop) (cddr prop))))
  )

; is-negated returns t if the proposition in negated, nil otherwise
(defun is-negated (prop) (and (listp prop) (equal (car prop) 'not)))

(defun logical-conn (pred) 
  (or (member pred '(and or implies ~> <=>)) 
      (numberp pred)
      (and (listp pred)
	   (numberp (car pred))
	   (<= (car pred) 1.0))))
(defun episodic-conn (pred) (member pred '(** * @ _)))
(defun causal-conn (pred) (member pred '(because)))

(defun is-lambda-pred (pred) (and (listp pred) (equal (car pred) 'L)))

(defun is-lambda-expr (pred)
  (cond ((null pred) nil)
	((equal (pred-type pred) 'lambda))
	((atom pred) nil)
	((equal (car pred) 'L) t)
	((listp (car pred)) (is-lambda-expr (car pred)))
	(t nil)))

(defun is-pred (p)
  (cond ((symbolp p) 
	 (or (pred-p p)
	     (and (epi-variable p)
		  (or (equal (find-sort p) 'pred)
		      (equal (find-sort (cadr (assoc p *subs*))) 'pred)))))
	((is-lambda-pred p) t)
	((and (listp p) (numberp (car p)) (<= (car p) 1.0)) t)
	((and (listp p) (is-pred (cadr p)) (is-operator (car p) 'pred)) t)
	;modified check needed here?
	((numberp p) t)
	(t nil))
  )

(defmacro is-function (f) `(function-p ,f))
(defmacro is-function-term (f) 
  `(and (listp ,f) (is-function (car ,f)) (cdr ,f)))

(defun is-quoted-expression (item)
  (and (listp item) (equal (car item) 'quote)))
(defun is-quasi-quoted-expression (item)
  (and (listp item) (equal (car item) 'qquote)))
(defun is-record (item)
  (and (listp item) (member (car item) *record-indicators*))
  )

(defun record-contents (item &aux 
			     (record (getp item 'print)))
  ; get rid of $
  (if (and (listp record) (member (car record) *record-indicators*))
      (setq record (cdr record)))  
  (if (and (listp record) (listp (car record)) 
	   (member (caar record) *quoted-indicators*) 
	  ; (member (cadr record) *sorts*) sorts may not be restricted to these
	   )
      (setq record (cdr record)))  ; get rid of sort
  record)

(defun is-wff (wff &aux isw)
  (or (and (epi-variable wff) 
	   (or (equal (find-sort wff) 'wff)
	       (equal (find-sort (cadr (assoc wff *subs*))) 'wff)))
  (and (listp wff)
       (not (is-function (car wff)))
       (or (is-quantifier (car wff))
	   (is-negated wff)
	   (setq isw (and (is-operator (car wff) 'wff t) (is-wff (cadr wff))))
	   (is-pred (prop-pred wff)))
       (if isw
	   t
	   (and (not (is-operator (car wff) 'pred))
		(not (is-operator (car wff) 'term))))
       ))
  )

; Sort stuff

(defun find-sort (item &aux op opt pr)
  (cond ((symbolp item)
	 (cond ((getp item 'sort))
	       ((function-p item) 'function)
	       ((operator-p item) 'operator)
	       ((quantifier-p item) 'quantifier)
	       ((pred-p item) 'pred)
	       ((wff-p item) 'wff)
	       ((equal (term-type item) 'record) 
		(setq pr (getp item 'print))
		(if (listp (cadr pr))
		    (cadadr (getp item 'print))))
	       (t nil)))
	((numberp item)
	 (if (integerp item) 'integer 'real))
	((stringp item) 'string)
	((is-quoted-expression item) nil)
	((is-quasi-quoted-expression item) nil)
	((operator-p (setq op (car item)))
	 (cond ((operator-result op))
	       ((setq opt (operator-type op)) (operator-result opt))
	       (t nil)))
	((equal (car item) 'L) 'pred)
	((is-wff item) 'wff)
	; special entry for functions?
	(t nil))
  )

; continued information routines 

(defun is-operator (op &optional result wff-force &aux res ot nr)
  (cond ((null op) nil)
        ((numberp op) t)
        ; no variables are allowed as sentence operators - should fix.
	((epi-variable-p op)
	 (let ((s (find-sort op)))
	      (cond ;((and (not wff-force) (equal result 'wff)) nil)
		    ((and wff-force (equal result 'wff) (equal s 'operator))
		     'operator)
		    ((equal result 'wff) nil)
		    ((equal s 'operator) 'operator)
		    ;((equal s 'wff-op) 
		    ; (if (equal result 'wff) 'operator nil))
		    ((equal (find-sort (cadr (assoc op *subs*))) 'operator)
		     'operator)
		    ((and result s) 
		     (if (member result '(quantifier))
			 (if (equal result s) 'operator nil)
			 nil))
		    (t nil))) )  ;????
        ((and (symbolp op) 
	      (or (operator-p op) (equal (find-sort op) 'operator)))
         ; return result - if unknown just return t
         (cond ((setq res (operator-result op))
                (if result (equal res result) t))
               ((setq ot (operator-type op))
                (cond ((setq res (operator-result ot))
                       (if result (equal res result) t))
                      (t t)))
               (t t)))
        ((atom op) nil)
        ((setq nr (is-operator (car op))) 
         (if result
             (if (symbolp (car op))
                 (or (and (equal (operator-result (car op)) result) 
			  (equal result 'pred))
		     (equal (operator-result (operator-result (car op))) result))
                 (or (equal nr 'operator) (equal nr result)))
             t))
        (t nil))
  )
 
(defun is-quantifier (item)
  (cond ((quantifier-p item) t)
	((atom item) nil)
        (t (is-operator (car item) 'quantifier))
        ))

(defun is-variable (var)
  (and (symbolp var)
       (or (epi-variable-p var)
	   (and (if (operator-p var) (equal (getp var 'sort) 'operator) t)
		(if (quantifier-p var) (equal (getp var 'sort) 'quantifier) t)
		(not (function-p var))
		(not (getp var 'indicators)) ) ) ))

(defun type-of-concept (arg)
  (cond ((listp arg) 'function)
        ((epi-variable-p arg) 'variable)
        (t 'constant))
  )


; Predefined Quantifiers 
(mapcar #'(lambda (x) (setp x 'QUANTIFIER t) x)
			 '(A E MOST MANY SOME FEW NOne WH THE E!))

(defun quantifier-prob (quant prob) (setp quant 'prob prob))

(quantifier-prob 'A 1)
(quantifier-prob 'E 1)
(quantifier-prob 'E! 1)
(quantifier-prob 'THE 1)
(quantifier-prob 'WH 1)
(quantifier-prob 'most 0.8)
(quantifier-prob 'many 0.6)
(quantifier-prob 'some 0.4)
(quantifier-prob 'few 0.2)
(quantifier-prob 'none 0.0)

(defun var-present (item &aux r)
  (cond ((null item) nil)
        ((not (symbolp item)) nil)
        ((epi-variable-p item) (not (getp item 'lambda)))
        ((and (setq r (getp item 'pred)) (var-present r)) t)
        ((and (setq r (getp item 'operator)) (var-present r)) t)
        ((and (setq r (getp item 'arguments))
              (eval (append (list 'or) (mapcar 'var-present r)))) t)
        (t nil))
  )

; number-arg returns t if arg is a concept of sort number or a number
 
(defun number-arg (arg)
  (cond ((symbolp arg)
	 (if (member (find-sort arg) '(number integer real))
	     t
	     nil))
	((atom arg) (numberp arg))
	(t nil))
  )

; non-constant-arg tests a list of arguments to see if any are not constants
; it returns t if any argument was not a constant, nil otherwise
 
(defmacro non-constant-arg (arglist)
  `(eval (append (list 'or) (mapcar 'var-present ,arglist)))
  )

(defun constant-arg (arg) (not (epi-variable arg)))

; check-constant checks to see if a given argument is a constant - a quoted
; expression or a constant concept is a constant, anything else is not

(defun check-constant (arg)
  (cond ((symbolp arg)
         (equal (getp arg 'term-type) 'constant))
        (t t))
  )

(defmacro get-min (hier node)
  `(cdr (assoc ,hier (getp ,node 'MIN))))
(defmacro get-id (hier node)
  `(cdr (assoc ,hier (getp ,node 'ID))))

(defmacro get-max (hier node)
  `(cdr (assoc ,hier (getp ,node 'MAX))))

(defmacro get-super (hier node)
  `(cdr (assoc ,hier (getp ,node 'SUPER))))

(defmacro get-sub (hier node)
  `(cdr (assoc ,hier (getp ,node 'SUB))))

(defmacro get-all-sub (node)
  `(remove 'nil (apply 'append (mapcar #'(lambda (a) (if (getp (car a) 'part-hier)
                                            nil (cdr a)))
                          (getp ,node 'SUB)))))
(defmacro get-all-super (node)
  `(remove 'nil (mapcar #'(lambda (a) (if (getp (car a) 'part-hier) nil (cdr a)))
                          (getp ,node 'SUPER))))
 

