;; January 2004 added the variable *print-to* to modify the stream to
;; which the printing functions should print. FM.

;; Last change August 1993
(in-package lib)


;
; IO and Error control routines
;

; tweakables here need to be available before tweakable defined
; These are redefined as tweakable in the tweak file to avoid interdependencies
(defvar *print* t) ;Indicates whether or not to print
(defvar *warn* t) ; indicates whether to print warning messages

(defvar *indent* 0 "current indenting amount")
(defvar *spaces* "" "current indent")

(defvar *print-to* nil)

 ;;;
 ;;; Miscellany to produce output.
 ;
(defun prinlis (list) (if *print* (mapc #'(lambda (w) (princ w *print-to*)) list)) nil)

(defconstant *nl* '#\newline)

(defun indent (num)
 ;;
 ;; print blanks for indentation.
 ;;
 (print-line)
 (if *print* (dotimes (j num) (princ " " *print-to*)))
 )

(DEFUN depth-indent (depth)
  (indent (* depth 2)))

(defun set-indent (depth &aux splist)
   (setq *indent* (* depth 2))
   (dotimes (i *indent*) (push " " splist))
   (setq *spaces* (squish splist))
   )

(defun reset-indent (spaces indent)
  (setq *spaces* spaces)
  (setq *indent* indent)
  )

(defun print-it (&rest stuff) (prinlis stuff))

(defun print-line (&rest stuff) (prinlis (append (list *nl* *spaces*) stuff)))

(defun p-error (stuff) (prinlis (append (list *nl* *spaces* "ERROR: ") stuff)))

(defun print-error (&rest stuff) (p-error stuff))

(defun print-warn (stuff) 
  (prinlis (append (list *nl* *spaces* "WARNING: ") stuff)))

;
; Warnings and errors
; 

(defun fatal-error (where &rest message)
  (throw where (p-error message)))

(defun warn-error (&rest message)
  (when *warn* (print-warn message))
  )
