;       Last change    August 1993
(in-package meta)

;                 EVALUATION Routines

(defun evaluate (netname lit negp pred arglist effort &aux rtn
			 (res 'unknown)) 
  (declare (ignore effort negp netname))
  (if (or (null arglist)
	  (not (symbolp pred))
	  )
      (setq res 'unknown)
      (let ()
	   (if (traced-p 'epilog::meta-eval)
	       (print-line "    Meta-specialist: Evaluating " (print-info lit)))
      ; find routine for the given pred and use that
      (setq res
	   (if (and (setq rtn (find-symbol (symbol-name pred) 'meta))
		    (not (or (not (fboundp rtn))
			     (macro-function rtn)
			     (special-operator-p rtn))))
	       (apply rtn arglist)
	       (if (null (cdr arglist))
	           (eval-for-prop (car arglist) pred)
		   'unknown)) )
	   (when (null res) (setq res 'unknown))
	   (when (and res (not (equal res 'unknown))
		      (traced-p 'epilog::meta-eval))
		 (print-line "    Meta-specialist: Evaluated " (print-info lit) 
			     " to " res))
	   ))
  res)

(defun active-topic (item &aux (ind (getp item 'indicators)))
  (setq ind (mapcar #'(lambda (a) (if (listp a) (car a) a)) ind))
  (dolist (i ind)
          (if (member i response-pkg::*active-topics* :test #'(lambda (a b)
                                (member (hier-rel a b) '(equivalent subsumed))))              (return-from active-topic t)))
  nil)

(defun default-info (prop pred)
  (declare (ignore pred))
  (case prop
	(%telic ; must refer to things which have distinct start and end
		)
	(%stative ; states
		  )
	(%nonvolitional
	 )
	;(%action-pred (active-topic pred))
	)
  )

(defun eval-for-prop (item prop &optional opsok allowed-ops &aux arg ap pt r)
  ; check to see if prop on item's symbol list  (mainly for checking preds)
  ; if item is a lambda pred, check it's predicate
  ; and/or - check all possibilities
  (cond ((null item) 'no)
	((not (symbolp item)) nil)
	((getp item prop) 'YES)
	((setq pt (getp item 'pred-type))
	 (case pt 
	       (lambda
		(if (and (setq ap (eval-for-prop (getp item 'main) prop opsok allowed-ops))
			 (not (equal ap 'unknown)))
		    ap))
	       ; not guaranteed - operator may change prop for predicate?
	       (modified 
		(cond ((and allowed-ops (member (getp item 'operator) allowed-ops))
		       'yes)
		      (opsok
		       (if (and (= (list-length (setq arg (getp item 'arguments))) 1)
		    
				(setq r (eval-for-prop (car arg) prop opsok allowed-ops))
				(equal r 'YES))
			   r
			   (eval-for-prop (getp item 'operator) prop opsok allowed-ops)))
		      (t 'unknown)))
	       (constant (default-info prop item))
	       (otherwise 'unknown)))
	((setq pt (getp item 'wff-type)) ;lambda
	 (case pt
	       (episodic (eval-for-prop (car (getp item 'arguments)) prop opsok allowed-ops))
	       (quantified 
		(if (and opsok (setq ap (eval-for-prop (getp item 'main) prop opsok allowed-ops))
			 (not (equal ap 'unknown)))
		    ap
		    'unknown))
	       (logical ; make sure all parts have that property
                (if opsok
		    (let ()
			 (dolist (l (getp item 'arguments))
			    (case (eval-for-prop l prop opsok allowed-ops) 
			      (no (return-from eval-for-prop 'no))
			      (unknown (return-from eval-for-prop 'unknown))
			      (yes nil)
			      (otherwise (return-from eval-for-prop 'unknown)))
			     )
			 'yes)
		    'unknown))
	       (modified 
		(if opsok
		    (eval-for-prop (car (getp item 'arguments)) prop opsok allowed-ops)
		    'unknown))
	       (prefix (eval-for-prop (getp item 'pred) prop opsok allowed-ops))
	       (otherwise 'unknown)))
	(t 'unknown))
  )

(defun %action-pred (&rest args) 
  (eval-for-prop (car args) 'epilog::%action-pred t)
  )

(defun %old-action-pred (&rest args &aux ap
			  (p (car args))
			  pt)
  (cond ((null p) 'no)
	((not (symbolp p)) nil)
	((getp p 'epilog::%action-pred) 'YES)
	((setq pt (getp p 'pred-type))
	 (case pt 
	       (lambda
		(if (and (setq ap (%action-pred (getp p 'main)))
			 (not (equal ap 'unknown)))
		    ap))
	       (modified (%action-pred (car (getp p 'arguments))))
	       (otherwise 'unknown)))
	((setq pt (getp p 'wff-type))
	 (case pt
	       (episodic (%action-pred (car (getp p 'arguments))))
	       (quantified 
		(if (and (setq ap (%action-pred (getp p 'main)))
			 (not (equal ap 'unknown)))
		    ap))
	       (modified (%action-pred (car (getp p 'arguments))))
	       (prefix (%action-pred (getp p 'pred)))
	       (otherwise 'unknown)))
	(t 'unknown))
  )

(defun action-type (&rest args &aux
			  (p (car args))
			  pt)
  (cond ((null p) 'NO)
	((not (symbolp p)) 'NO)
	((setq pt (getp p 'term-type))
	 (case pt
	       (modified
		(if (equal (getp p 'operator) 'epilog::to)
		    (%action-pred (car (getp p 'arguments)))
		    'YES) )  ;?
	       (otherwise 'unknown)) )
	(t 'unknown))
  )

(defun %type-predicate (&rest args &aux
			  (p (car args)))
  (cond ((null p) 'NO)
	((not (symbolp p)) 'NO)
	((type-pred-p p) 'YES)
	(t 'NO))
  )

(defun %formula (&rest args &aux
			  (p (car args)))
  (cond ((null p) 'NO)
	((not (symbolp p)) 'NO)
	((equal (getp p 'entity-type) 'wff) 'YES)
	(t 'unknown))
  )

(defun %action-formula (&rest args &aux
			  (wff (car args))
			  wt)
  (cond ((null wff) 'NO)
	((not (symbolp wff)) 'NO)
	((setq wt (wff-type wff))
	 (case wt 
	       (quantified 'No)
	       ((episodic prefix) (%action-pred wff))
	       (otherwise 'NO)))
	(t 'NO))
  )

;;  *** %event formula -- needs to be updated!!!!
(defun %event-formula (&rest args &aux
			  (wff (car args))
			  wt)
  (cond ((null wff) 'NO)
	((not (symbolp wff)) 'NO)
	((setq wt (wff-type wff))
	 (case wt 
	       (quantified 'NO)
	       ; as long as it doesn't describe a process or state
	       ; then true, else false
	       (modified (%event-formula (car (getp wff 'arguments))))
	       (episodic 'YES)
	       (prefix 'YES) 
	       (otherwise 'NO)))
	(t 'NO))
  )

; these need to be updated properly!!!!!
; stative - check indicators on predicates??
(defun %stative (&rest args) 
  (eval-for-prop (car args) 'epilog::%stative t))

; telic - has distinct start and end - how to test for that here?
(defun %telic (&rest args) 
  (eval-for-prop (car args) 'epilog::%telic t)
  )

(defun %nonvolitional (&rest args) 
  (eval-for-prop (car args) 'epilog::%nonvolitional t)
  )

(defun %unlocated (&rest args) 
  (eval-for-prop (car args) 'epilog::%unlocated t)
  )

(defun %contains (&rest args &aux
			(item (car args))
			(lookfor (cadr args))
			(res 'unknown))
  (when (and item lookfor (null (cddr args)))
	; Look for lookfor in item - see if it occurs at ANY level of embedding
	(setq item (print-info item))
	(if (member (car item) *quoted-indicators*)
	    (setq item (cadr item)))
	(setq lookfor (print-info lookfor))
	(if (member (car lookfor) *quoted-indicators*)
	    (setq lookfor (cadr lookfor)))
	(if (meta-contains lookfor item)
	    (setq res 'YES)
	    (setq res 'NO))
	)
  res)

(defun meta-contains (lookfor item)
  (cond ((null item) nil)
	((equal lookfor item) t)
	((atom item) nil)
	(t (dolist (i item)
		   (if (meta-contains lookfor i)
		       (return-from meta-contains t))
		   ))
	)
  )

(defun %numeral-term (&rest args &aux
			    (a1 (car args)))
  (when (and a1 (null (cdr args)))
	(if (numberp a1) 'YES 'NO))
  )

(defun %episodic-term (&rest args &aux s
			     (a1 (car args)))
  (when a1 (setq s (find-sort a1)))
  (when (and a1 (null (cdr args)))
	(setq s (find-sort a1))
	(if (member-if #'(lambda (a) 
				 (member (hier-rel a s) '(equivalent subsumes)))
		       '(episode e-term))
	    'YES
	    'NO)
	)
  )

(defun %atemporal-formula (&rest args &aux 
			  (p (car args)))
  (when (and p (null (cdr args)))
	(is-atemporal p))
  )

(defun is-atemporal (wff &aux ans)
  (case (wff-type wff)
	(episodic 'YES)
	(causal 'YES)
	(logical 
	 (dolist (a (getp wff 'arguments))
		 (setq ans (is-atemporal a))
		 (if (equal ans 'NO) (return-from is-atemporal 'NO))
		 (if (equal ans 'unknown) (return-from is-atemporal 'unknown))
		 )
	 'YES)
	(prefix
	 (let ((pred (getp wff 'pred)))
	      (cond ((equal pred 'equal) 'YES)
		    ((type-pred-p pred) 'YES) ; incorrect, but leave for now
		    ; better test needed here
		    ((member-if #'(lambda (a) 
			  (member (if (listp a) (car a) a) 
			    '(epilog::tp.time-relationship
			      epilog::tp.causal-relationship)))
				(getp pred 'epilog::indicators)) 'YES)
		    (t 'NO))))
	(quantified (is-atemporal (getp wff 'main)))
	(quasi-quoted-expression (is-atemporal (getp wff 'quote-wff)))
	(modified (is-atemporal (car (getp wff 'arguments))))
	(otherwise 'NO)
	))

(defun %*-or-** (&rest args &aux 
		       (p (car args)))
  (when (and p (null (cdr args)))
	(if (member p '(* **)) 'YES 'NO))
  )

(defun %@ (&rest args &aux 
		       (p (car args)))
  (when (and p (null (cdr args)))
	(if (equal p '@) 'YES 'NO))
  )

; kind level - you can say p is widespread
(defun %kind-level (&rest args)
  (eval-for-prop (car args) 'epilog::%kind-level nil '(K Ka Ke))
  )

(defun %object-level (&rest args)
  ; specified predicates like drink, see, eat, etc
  ; also (LzLy(y (make L x(x=z)))) 
  (eval-for-prop (car args) 'epilog::%object-level nil '(epilog::make))
  )

(defun %currently-I-cannot-infer-likely (&rest args &aux res
				    (qa-iter epilog::*qa-iterations*))
  ; only true of specific instant evaluated at 
  ; do a small inference attempt on arg1
  ; if it does not return YES, or it returns NO, then this function is YES
  (cond ((or epilog::*meaning-postulate* epilog::*simplification-schema*
	     (not (or epilog::*ok-to-verify* epilog::*qa-mode*))
	     ;(getp (getp (car args) 'epilog::variable) 'epilog::matchable)
	     (epi-variable (car args)))
	 'unknown)
	(t
  (tweak 'epilog::*qa-iterations* 10)
  (setq epilog::*key-verify* t)
  (setq res (car (epilog::question (getp (car args) 'print) nil 2 " Evaluating likelihood " nil)))
  (epilog::qa-cleanup nil)
  (tweak 'epilog::*qa-iterations* qa-iter)
  (setq epilog::*key-verify* nil)
  (case (car res)
	(YES 'NO)
	(NO 'YES)
	(otherwise 'YES)) ) )
  )

(defun %kind-of-episode (&rest args &aux
			       (item (car args)))
  (cond ((null item) ''NO)
	((not (symbolp item)) 'NO)
	((epi-variable item) 'unknown)
	((equal (getp item 'entity-type) 'term)
	 (case (getp item 'term-type)
	       (modified
		(if (equal (getp item 'operator) 'epilog::Ke)
		    'YES
		    (if (epi-variable (getp item 'operator))
			'unknown
			'NO)))
	       (otherwise 'NO)))
	(t 'NO))
  )

