;       Last change    August 1993
(in-package number)

;                         FUNCTIONS
; These functions are functions the user may have in propositions
; When evaluated, they use the global specialist effort level for effort

; get-subnet determines the apporpriate number subnet to use

(defun get-subnet (net)
  (if (null net)
      (let ()
	   (if (null current-net)
	       (setq current-net (determine-net *current-netname*)))
           current-net
           )
      (determine-net net))
  )

; find-max finds the maximum of a named number

(defun find-max (arg net &aux
		     (nbr (findnumber net arg)))
  (cond ((null nbr) nil)
        ((epinumber-max nbr))
        (t nil))
  )

; max-arg returns the maximum of an agument - if it is a named number, its max 
; bound is returned. If a quoted expression, its value

(defun max-arg (arg net)
  (unless (null arg)
          (if (numberp arg)
	      arg
	      (if (and (symbolp arg) (equal (getp arg 'term-type) 'record))
		  arg
              (find-max arg net)))
	  )
  )

; find-min finds the minimum of a named number

(defun find-min (arg net &aux
		     (nbr (findnumber net arg)))
  (cond ((null nbr) nil)
        ((epinumber-min nbr))
        (t nil))
  )

; min-arg returns the minimum of an agument - if it is a named number, its max 
; bound is returned. If a quoted expression, its value

(defun min-arg (arg net)
  (unless (null arg)
          (if (numberp arg)
              arg
	      (if (and (symbolp arg) (equal (getp arg 'term-type) 'record))
		  arg
              (find-min arg net)))
	  )
  )

; find-value returns the value of a named number - if it has no exact value,
; a list of its minimum and maximum are returned

(defun find-value (arg net &aux
		     (nbr (findnumber net arg)))
  (cond ((null nbr) nil)
        ((epinumber-value nbr))
	(nil ; (and (epinumber-max nbr)
	    ; (epinumber-min nbr))
	 (list (epinumber-min nbr) (epinumber-max nbr)))
        (t nil))
  )

; value arg returns the value (min and max) of a named number, or the value of
; a quoted expression

(defun value-arg (arg net)
  (unless (null arg)
          (if (numberp arg)
	      arg
	      (if (and (symbolp arg) (equal (getp arg 'term-type) 'record))
		  arg
              (find-value arg net))))
  )

; max-of returns the maximum  of all its arguments. If any are named numbers,
; their maximum bound is used.

(defun max-of (&rest args &aux
		       (usenet (get-subnet nil))
		       res newargs)
  ; for any argument that is a concept, replace it by its maximum
  ; then take maximum of all the results

  (unless (null args)
	  (setq newargs (mapcar #'(lambda (a) (max-arg a usenet)) args))
	  (when (eval (append (list 'and) (mapcar #'numberp newargs)))
	        (setq res (apply #'max newargs))
	        )
	  )
  (if (or (traced-p 'epilog::function-eval)
	  (traced-p 'epilog::number-function-eval))
      (print-line "    Number specialist: MAX-OF " args " evaluated to " 
		  (print-info res)))
  res)

; min-of returns the minimum  of all its arguments. If any are named numbers,
; their minimum bound is used.


(defun min-of (&rest args &aux
		       (usenet (get-subnet nil))
		       res newargs)
  ; for any argument that is a concept, replace it by its minimum
  ; then take minimum of all the results

  (unless (null args)
	  (setq newargs (mapcar #'(lambda (a) (min-arg a usenet)) args))
	  (when (eval (append (list 'and) (mapcar #'numberp newargs)))
	        (setq res (apply #'min newargs))
	        )
	  )
  (if (or (traced-p 'epilog::function-eval)
	  (traced-p 'epilog::number-function-eval))
      (print-line "    Number specialist: MIN-OF " args " evaluated to " 
		  (print-info res)))
  res)

; value-of returns the exact value a number has - if it has no exact value,
; a list of the minimum and maximum is returned

(defun value-of (&rest args &aux
		       (concept (car args))
		       (usenet (get-subnet nil))
		       res newargs)

  (unless (null concept)
	  (setq newargs (value-arg concept usenet))
	  (if (equal (car newargs) (cadr newargs)) (setq res (cadr newargs)))
	  )
  (if (or (traced-p 'epilog::function-eval)
	  (traced-p 'epilog::number-function-eval))
      (print-line "    Number specialist: VALUE-OF " args " evaluated to " 
		  (print-info res)))
  res)

; add returns the result of adding all arguments

(defun add (&rest args &aux
		       (usenet (get-subnet nil))
		       res newargs)
  ; for any argument that is a concept, replace it by its value
  ; then add of all the results

  (unless (null args)
	  (setq newargs (mapcar #'(lambda (a) (value-arg a usenet)) args))
	  (when (eval (append (list 'and) (mapcar #'numberp newargs)))
	        (setq res (apply #'+ newargs))
	        )
	  )
  (if (or (traced-p 'epilog::function-eval)
	  (traced-p 'epilog::number-function-eval))
      (print-line "    Number specialist: ADD " args " evaluated to " 
		  (print-info res)))
  res)

; relation returns the most strict relation between its arguments

(defun relation (&rest args &aux
		       (a1 (car args))
		       (a2 (cadr args))
		       (usenet (get-subnet nil))
		       res)
   (unless (or (null a1) (null a2))
           (if (or (not (number-arg a1)) (not (number-arg a2)))
	       nil  ; not applicable to this specialist
	       (let ()
		    (setq res
                           (if (and (symbolp a1) (symbolp a2))
		               (find-relation usenet a1 a2
			    		      *specialist-eval-effort*)
		               (if (symbolp a1)
			           (get-nbr-reln usenet a1 a2)
			           (if (symbolp a2)
			               (reverse-reln (get-nbr-reln usenet 
				    	    a2 a1))
			               (cp-nbrs a1 a2))
			           )))
		    (setq res (find-symbol (symbol-name res) 'epilog))
		    (if (or (traced-p 'epilog::function-eval)
			    (traced-p 'epilog::number-function-eval))
                         (print-line "    Number Specialist: RELATION " 
					  args " evaluated to " 
					  (print-info res)))
		    )))
  res)

