;       Last change    August 1993
(in-package set)

; the following routines are used to determine whether two literals are 
; compatible or incompatible
; No residues are returned. It is possible that actions are missed by not
; investigating using residues, but it is still unclear whether comparison
; using this specialist helps anyway. Probably it should be changed to use
; a final comparison after temporary entry that is more like the time specialist
; uses.


; classify-lits classifies the given literals, and returns a list
; of substitution orders and which literal to enter.
; Each item in the list is of the form (orderlist . number of lit to enter)
; Each item in orderlist is of the form 
;     (element number from 1st lit . element number from second)

(defun classify-lits (neg-p1 pred1 neg-p2 pred2 arglist1 arglist2 &aux which)
  (if neg-p1 (setq which 2))
  (if neg-p2 (setq which 1))
  (unless which
	  (if (equal pred1 'member-of-0) (setq which 2))
	  (if (equal pred2 'member-of-0) (setq which 1)) )
  (unless which
	  (if (member-if 'immed-set arglist2)
	      (setq which 1)
	      (if (member-if 'immed-set arglist1)
		  (setq which 2)
		  (setq which 1))))

  ; each literal has only 2 args which can be unified
  ; only two possible unifications - same direction and reversed
  (cond ((member pred1 '(member-of member-of-0))
	 (if (member pred2 '(member-of member-of-0 subset-of)) 
	    (list (cons '((1 . 1) (2 . 2)) which))
	    (list (cons '((1 . 1) (2 . 2)) which)
		  (cons '((1 . 2) (2 . 1)) which))))
	((equal pred1 'equal) (list (cons '((1 . 1) (2 . 2)) which)
				    (cons '((1 . 2) (2 . 1)) which)))
	(t (if (member pred2 '(member-of member-of-0)) 
	    (list (cons '((1 . 1) (2 . 2)) which))
	    (list (cons '((1 . 1) (2 . 2)) which)
		  (cons '((1 . 2) (2 . 1)) which))))
	)
  )

; incompatible-add checks to make sure that the residues are not the same as the
; original literals (without substitutions). If so, a resolving item is
; added, otherwise, nothing is done.

(defun incompatible-add (lit-1 lit-2 res1 res2 eval1 eval2 subs1 subs2
			       &optional (incompatible t)
			       &aux rr)
  (declare (ignore eval1 eval2))
  (setq res1 (substitute-prop (normalize-wff res1) subs1))
  (setq res2 (substitute-prop (normalize-wff res2) subs2))
  (unless (or (equal (print-info lit-1) res1)
	      (equal (print-info lit-1) res2)
	      (equal (print-info lit-2) res1)
	      (equal (print-info lit-2) res2))
          (when (traced-p 'epilog::set-test)
		(print-line "    Set Specialist: Found " (print-info lit-1 )
		     (if incompatible " incompatible" " compatible") " with "
		     (print-info lit-2) )
		(print-line "       with residues: "
			    (if (setq rr (append res1 res2)) rr "null"))
		(print-line "       with substitutions: " 
			    (print-subs (append subs1 subs2))))
	  (list 
	   (make-comparison-info :lit1 lit-1 :lit2 lit-2
			     :subs1 subs1 :subs2 subs2 
			     :residue1 res1
			     :residue2 res2
			     ))
	  )
  )


; create-residue creates a literal of the given neg pred ard args

(defun create-residue (neg pred &rest args &aux
			       (reln (find-symbol (symbol-name pred) 'epilog)))
  (if (or (null pred)
	  (null args)
	  (eq (car args) (cadr args)))
      t  
      (make-prop neg :pred reln :args args)
      )  
  )

; enter-to-test just ensuere that no negations get passed to enter

(defun enter-to-test (net negp pred arglist)
  (if (or negp (equal pred 'member-of-0))
      nil 
      (enter-the-set-info net negp pred arglist))
  )


(defun incompatible-enter-and-compare (net lit-1 lit-2 entry eval1 eval2
                                      neg-p1 reln1 newargs1
                                      neg-p2 reln2 newargs2 subs1 subs2 effort
                                      &optional (incompatible t)
                                      &aux
                                      res cp resans)
  ; see if a membership test (both preds member-of )
  (if (and (equal (car newargs1) (car newargs2))
	   (or (member-test reln1 reln2)
	       (subset-test reln1 reln2 newargs1 newargs2)))
      (setq res (test-membership incompatible net lit-1 neg-p1 (car newargs1) reln1 
			   (cadr newargs1) lit-2 neg-p2 (car newargs2) reln2 
			   (cadr newargs2) subs1 subs2 effort)))
  (if (and (equal (cadr newargs1) (cadr newargs2))
	   (subset-test reln1 reln2 newargs1 newargs2))
	  (setq res (append res (test-subset-2 incompatible net lit-1 neg-p1 
					       (car newargs1) lit-2 neg-p2 
					       (car newargs2) subs1 subs2 
					       effort)))
      )
  (unless (or res (null entry))
	  (if (and (equal eval1 'yes) (not neg-p1) (equal entry 2)) (setq entry 1))
	  (if (and (equal eval2 'yes) (not neg-p2) (equal entry 1)) (setq entry 2))

  (setq cp (checkpoint "Set-CP"))
  (when (and (equal entry '2) (not (equal eval2 'no))
	     (not (member eval1 '(YES NO))))
        (enter-to-test net neg-p2 reln2 newargs2)
        (setq resans (eval-with-neg net neg-p1 reln1 newargs1 effort))
	(cond ((setq res
		     (incompatible-with-eval lit-1 lit-2 resans eval2 subs1 subs2
					     (traced-p 'epilog::set-test)
		                        "    Set Specialist:" incompatible))
		 res)
        ))
  (when (and (equal entry '1) (not (equal eval1 'no))
	     (not (member eval2 '(YES NO))))
        (enter-to-test net neg-p1 reln1 newargs1)
        (setq resans (eval-with-neg net neg-p2 reln2 newargs2 effort))
	(setq res (incompatible-with-eval lit-1 lit-2 eval1 resans subs1 subs2
					     (traced-p 'epilog::set-test)
		                        "    Set Specialist:" incompatible))
	)
	  (retract cp :remove t)
	  )
  res)

(defun useful-args (a b)
  (and (or (and (record-p a) (equal (find-sort a) 'set))
	   (findset current-net a))
       (or (and (record-p b) (equal (find-sort b) 'set))
	   (findset current-net b)))
  )

; test-residue does the necessary substitutions, and then resolves or
; factors the resulting literals.

(defun test-residue (incompatible net lit-1 lit-2 ord-entry
			     neg-p1 reln1 arglist1
			     neg-p2 reln2 arglist2 effort &aux
			     (ordlist (car ord-entry))
			     (entry (cdr ord-entry))
			     (newargs1 arglist1)
			     (newargs2 arglist2)
			     (eval1 'unknown)
			     (eval2 'unknown)
			     newlit1 newlit2
	                     res ulist subs1 subs2)
  (unless (null ordlist)
	  ; unify the two literals according to specification given
          (setq ulist (unify-lists arglist1 arglist2 ordlist 'useful-args))

	  ; make substitutions in the two literals
          (unless (null ulist)
                  (setq subs1 (car ulist))
	          (setq subs2 (cdr ulist))
	          (setq newlit1 (normalize-wff (substitute-prop lit-1 subs1)))
	          (setq newlit2 (normalize-wff (substitute-prop lit-2 subs2)))
		  ))

  (unless (or (null newlit1)
	      (null newlit2))
	   (when (traced-p 'epilog::set-test)
		 (print-line "        Set Specialist: Comparing " 
			     (getp newlit1 'print)
			     " with " (getp newlit2 'print)))
          
	  ; evaluate the arguments 
	  (setq newargs1 (spec-eval-args newlit1))
	  (setq newargs2 (spec-eval-args newlit2))
	  (when (or (equal (car newargs1) (cadr newargs1))
		    (equal (car newargs2) (cadr newargs2)))
		(return-from test-residue nil))

	  ; evaluate the literals
	  (if (getp newlit1 'epilog::prob)
	      (progn (setq eval1 'YES)
		     (when (traced-p 'epilog::set-eval)
			   (print-line "    Set Specialist: Evaluated "
				       (getp newlit1 'print) " to " eval1)))
	      (setq eval1 (eval-with-neg net neg-p1 reln1 newargs1 effort)) )
	  (if (getp newlit2 'epilog::prob)
	      (progn (setq eval2 'YES)
		     (when (traced-p 'epilog::set-eval)
			   (print-line "    Set Specialist: Evaluated "
				       (getp newlit2 'print) " to " eval2)))
	      (setq eval2 (eval-with-neg net neg-p2 reln2 newargs2 effort)))

	  ; make sure substitution list complete
          (when (equal eval1 'unknown)
                (setq subs1 (union subs1 
		                (arg-sub-list newargs1 
						  (getp newlit1 'arguments) 1)))
	        (setq newlit1 (normalize-wff (substitute-prop lit-1 subs1)))
	        )
          (when (equal eval2 'unknown)
                (setq subs2 (union subs2 
		              (arg-sub-list newargs2 
						 (getp newlit2 'arguments) 2)))
	        (setq newlit2 (normalize-wff (substitute-prop lit-2 subs2)))
	        )

	  ; then use evaluations if possible

	  (cond ((setq res
		     (incompatible-with-eval lit-1 lit-2 eval1 eval2 subs1 subs2
					     (traced-p 'epilog::set-test)
		                        "    Set Specialist:" incompatible))
		 res)
		((and (member eval1 '(YES NO)) (member eval2 '(YES NO))) nil)

		((equal (if (equal entry 1) eval1 eval2) 'no) nil)

		((or (equal eval1 'NO) (equal eval2 'NO)) nil)

		(t (incompatible-enter-and-compare net
                         lit-1 lit-2 entry eval1 eval2 
			 neg-p1 reln1 newargs1
			 neg-p2 reln2 newargs2
			 subs1 subs2 effort incompatible))
		)
	  )
  )

; if arg is a constant, it must be of correct sort - if variable, must be
; of correct sort, or have no sort

(defun okarg (arg &aux
		  (s (find-sort arg)))
  (if (check-constant arg)
      (if (equal s 'epilog::set)
	  t
	  nil)
      (if (and (symbolp arg)
	       (member (getp arg 'term-type) '(variable record)))
	  (or (null s)
	      (if (equal s 'epilog::set)
		  t
	          nil))
	  nil)
      )
  )

; check-arg-sorts checks the arguments of the literal and determines if
; this specialist can resolve or factor with them

(defun check-arg-sorts (pred arglist)
  (case pred 
	(equal t)
	((member-of member-of-0) (okarg (cadr arglist)))
	(subset-of (and (okarg (car arglist)) (okarg (cadr arglist))))
	)
  )

; get-mems returns a list of members for the given argument. If recurs is
; true, it recursively replaces membrs which are sets with their members

(defun get-mems (net arg recurs)
  (if (not (get-set-immed arg))
      (unless (or (equal (getp arg 'term-type) 'epilog::function)
		  (not (findset net arg)))
      (if recurs
	  (get-contents-0 net arg)
	  (get-contents net arg)))
      (if recurs
	  (get-set-immed-0 net arg)
	  (get-set-immed arg))
      )
  )

; incompatible-lits tries to determine whether lit1 and lit2 are resolvable
; it tries all possible substitutions to find out

(defun incompatible-lits (netname lit1 lit2 neg-p1 pred1 arglist1 
		   	   neg-p2 pred2 arglist2 effort 
			   &optional (incompatible t) &aux
		           (net (determine-net netname))
		           (reln1 (find-symbol (symbol-name pred1) 'set))
		           (reln2 (find-symbol (symbol-name pred2) 'set))
			   (all-constants (not (or
					   (non-constant-arg arglist1)
					   (non-constant-arg arglist2))))
			   )

  (when (and (check-arg-sorts reln1 arglist1)
	     (check-arg-sorts reln2 arglist2))

  ; see if arguments unifiable, and in what order

  (if (traced-p 'epilog::set-test)
      (print-line "    Set Specialist: Trying to determine "
		  (if incompatible "incompatibility" "compatibility") " of " 
		  (print-info lit1) " and " (print-info lit2))
      (if (not (equal (episet-subnet-identifier net) '/))
	  (print-it "    in subnet " (episet-subnet-identifier net)))
      )

  (if all-constants
      (let ((eval1 (eval-with-neg net neg-p1 reln1 arglist1 effort))
	    (eval2 (eval-with-neg net neg-p2 reln2 arglist2 effort)))
	   ; try evaluations first
	   ;unless
	    (cond ((incompatible-with-eval lit1 lit2 eval1 eval2 nil nil
				      (traced-p 'epilog::set-test)
				      "    Set Specialist:" incompatible))
		   ; if no result there - see if a membership test - otherwise
		   ; cnanot do anything
		     ((and (equal (car arglist1) (car arglist2))
			       (member-test reln1 reln2))
		       (test-membership incompatible net lit1 neg-p1
				   (car arglist1) reln1 (cadr arglist1) lit2
				   neg-p2 (car arglist2) reln2 (cadr arglist2)
				   nil nil effort))
			 ((subset-test reln1 reln2 arglist1 arglist2)
			  (if (equal (car arglist1) (car arglist2))
			       (test-membership incompatible net lit1 neg-p1
				(car arglist1) reln1 (cadr arglist1) lit2
			         neg-p2 (car arglist2) reln2 (cadr arglist2)
				nil nil effort))
			  (if (equal (cadr arglist1) (cadr arglist2))
			      (test-subset-2 incompatible net lit1 neg-p1
					   (car arglist1) lit2 neg-p2 
					   (car arglist2) 
					   nil nil effort)))
			     )
		   )


      ; if not constants
      (if (and (not (equal reln1 reln2))
	       (or (equal reln1 'has-cardinality)
		   (equal reln2 'has-cardinality)))
	  nil
	  (let* ((ord-entry (classify-lits neg-p1 reln1 neg-p2 reln2 arglist1 arglist2))
		 r)
		(mapcar #'(lambda (o) (setq r (append r
		      (test-residue incompatible net lit1 lit2 o 
				    neg-p1 reln1 arglist1
				    neg-p2 reln2 arglist2 effort))))
			ord-entry)
		r)
	    )
      ))
  )

; compatible-lits tries to determine whether lit1 and lit2 are compatible.
; it tries all possible substitutions to do this.

(defun compatible-lits (netname lit1 lit2 neg-p1 pred1 arglist1 
		   	   neg-p2 pred2 arglist2 effort)
  (incompatible-lits netname lit1 lit2 neg-p1 pred1 arglist1 neg-p2 pred2
		     arglist2 effort nil)
  )

; member-test returns t if both predicates are membership predicates, nil
; otherwise 

(defun member-test (reln1 reln2)
  (and (member reln1 '(member-of member-of-0))
       (member reln2 '(member-of member-of-0)))
  )

; subset-test returns t if both predicates are subset predicates, and we can
; do something with them

(defun subset-test (reln1 reln2 arglist1 arglist2)
  (declare (ignore arglist1 arglist2))
  (and (equal reln1 'subset-of)
       (equal reln2 'subset-of)
;       (or (and (not (immed-set (car arglist1))) 
;		(not (immed-set (car arglist2))))
;	   (and (not (immed-set (cadr arglist1))) 
;		(not (immed-set (cadr arglist2)))))
       )
  )


; test-subset-2 determines the residue for two subset literals
;   union of members subset-of x

(defun test-subset-2 (incompatible net lit-1 neg-p1 ss1 lit-2 neg-p2 
				   ss2 subs1 subs2 effort
			     &aux
			     (setmems1 (get-mems net ss1 t))
			     (setmems2 (get-mems net ss2 t))
			     result use-lit)

  (declare (ignore effort))
  (setq result (if (not incompatible)
		   (cond ((and neg-p1 neg-p2)
			  (calc-union (list setmems1 setmems2)))
			 (neg-p1
			  (set-difference setmems1 setmems2))
			 (neg-p2
			  (set-difference setmems2 setmems1))
			 (t (calc-union (list setmems1 setmems2))))
		   (cond ((and neg-p1 neg-p2)
			  (calc-union (list setmems1 setmems2)))
			 (neg-p1
			  (set-difference setmems1 setmems2))
			 (neg-p2
			  (set-difference setmems2 setmems1))
			 (t (calc-intersect (list setmems1 setmems2))))))
  (if incompatible 
      ; if null result, null residue for resolving
      (unless (null result)
	  ;otherwise residue is result subset-of first arg (possibly negated)
	  (incompatible-add lit-1 lit-2
		       (substitute-prop (setq use-lit
					      (if neg-p2
						  lit-2
					          lit-1))
					(list 
					 (list (car (getp use-lit 'arguments))
					       (new-record result 'set)
					)
					       ))
		       nil nil nil
		       subs1 subs2))


      ; for factoring - if null result, no factoring action
      ; otherwise factor is result subset-of arg1 (possibly negated)
      (unless (null result) 
              (incompatible-add lit-1 lit-2  
		       (substitute-prop (setq use-lit
					      (if (and neg-p1 (not neg-p2))
						  lit-2
					          lit-1))
					(list 
					 (list (car (getp use-lit 'arguments))
					       (new-record result 'set)
					       )
					       ))
				nil nil nil
		  subs1 subs2 nil))

      )
  )

; test-membership determines whether two literals are resolvable or
; factorable and the resulting residues or factors, and adds these
; actions to the agenda  (x member-of intersection)

(defun test-membership (incompatible net lit-1 neg-p1 newm1 reln1 set1
				lit-2 neg-p2 newm2 reln2 set2 
				subs1 subs2 effort &aux 
				result setmems1 setmems2 use-lit)
  (declare (ignore effort newm1 newm2))
  (setq setmems1 (get-mems net set1 (equal reln1 'member-of-0)))
  (setq setmems2 (get-mems net set2 (equal reln2 'member-of-0)))


  ; result is calculated for the residue or factor - if null, no residue
  (setq result (if (not incompatible) ; prev - just incompatible
		   (cond ((and neg-p1 neg-p2)
			  (calc-union (list setmems1 setmems2)))
			 (neg-p1
			  (set-difference setmems2 setmems1))
			 (neg-p2
			  (set-difference setmems1 setmems2))
			 (t (calc-intersect (list setmems1 setmems2))))
		   (cond ((and neg-p1 neg-p2)
			  (calc-intersect (list setmems1 setmems2)))
			 (neg-p1
			  (set-difference setmems2 setmems1))
			 (neg-p2
			  (set-difference setmems1 setmems2))
			 (t (calc-union (list setmems1 setmems2))))))

  (if incompatible 
      ; if null result, null residue for resolving
      (if (null result)
          (incompatible-add lit-1 lit-2 nil nil nil nil subs1 subs2) 

	  ;otherwise residue is first arg member-of result (possibly negated)
	  (when (not (+member result))
	  (incompatible-add lit-1 lit-2
		       (substitute-prop (setq use-lit
					      (if (and neg-p1 (not neg-p2))
						  lit-2
					          lit-1))
					(list 
					 (list (cadr (getp use-lit 'arguments))
					       (new-record result 'set)
					       )
					       ))
		       nil nil nil
		       subs1 subs2)))


      ; for compatible - if null result, no action
      ; otherwise arg1 member-of result (possibly negated)
      (unless (or (null result) 
		  (+member result))
              (incompatible-add lit-1 lit-2
		       (substitute-prop (setq use-lit
					      (if (and neg-p1 (not neg-p2))
						  lit-2
					          lit-1))
					(list 
					 (list (cadr (getp use-lit 'arguments))
						     (new-record result 'set)
					       )
					       ))
				nil
				nil nil
		  subs1 subs2 nil))

      )
  )
