;       Last change    August 1993
(in-package set)


; add-member adds a member mem to set setname

(defun add-member (net setname mem  &optional (already nil) &aux
		       (curset (getset net setname))
		       (curmems (episet-contents curset))
		       newset)
  ; if this is the first member, make sure we flag the set as partially
  ; known by adding a + to its members
  (if (null curmems) (push 'epilog::+ curmems))
  (unless (member mem curmems)
	  (setq curset (writeable-copy net curset))
          (setq newset (pushnew mem curmems))
	  (if (traced-p 'epilog::set-entry)
	      (print-line "    Set Specialist: Adding member " mem 
			       " to set " setname))
	  (update-card net curset newset t)
	  (setq already (append already (list (findset net setname))))
	  (dolist (c (episet-connections curset))
		  (add-member-connect net c mem nil already))
	  (dolist (c (episet-connections-from curset))
		  (add-member-connect net c mem t already))
	  )
  )

(defun add-member-connect (net conn mem from already &aux
				(how (episet-connect-how conn))
				(sets (episet-connect-sets conn))
				(main-set (episet-connect-main-set conn))
				st st2)
  ; if from is nil, main-set is the one we just added to, else its the one
  ; we're going to add to
  (cond (from (unless (member main-set already)
		 (case how
		       (intersect 
			; check to see if new member in other set of
			; intersection; add if so
			(if (and (setq st (findset net (car sets)))
				 (equal (is-a-member mem st (episet-contents st) 
						     (episet-finished st)) t)
				 (setq st2 (findset net (cadr sets)))
				 (equal (is-a-member mem st2 (episet-contents st2) 
						     (episet-finished st2)) t)
				 )
			    (add-member net main-set mem already)) )
		       (union-of 
			; add new member
			(add-member net main-set mem already))
		       (set-diff 
			; if adding to first , check to see if not in 2nd
			(if (and (setq st (findset net (car sets)))
				 (equal (is-a-member mem st (episet-contents st) 
						     (episet-finished st)) t)
				 (setq st2 (findset net (cadr sets)))
				 (null (is-a-member mem st2 (episet-contents st2) 
						     (episet-finished st2)))
				 )
			    (add-member net main-set mem already)) )
		       )))
	(t (case how
		 (intersect ; add new member both to s2 and s3
                  (dolist (s sets)
			  (unless (member s already)
				  (add-member net s mem already))))
		 ((set-diff subset-of)
		  (unless (member (car sets) already)
			  (add-member net (car sets) mem already))))
	   ))
  )


; add-set sets the contents of a set setname to the members specified in
; a quoted set - given by setdesc

(defun add-set (net setname setdesc &aux
		    (curset (writeable-copy
			     net (getset net setname)))
		    (curmems (episet-contents curset))
		    newset)
  (if (equal curmems '(+)) (setq curmems nil))
  (check-equal-members net curmems setdesc)
  (setq newset (union curmems setdesc))
  (update-card net curset newset t)
  )

(defun enter-equal (net a1 a2)
  (cond ((or (equal (find-sort a1) 'set) ;(compatible-sorts 'set (find-sort a1))
	     (equal (find-sort a2) 'set) ;(compatible-sorts 'set (find-sort a2))
	     )
	 (enter-set net a1 a2))
	)
  )

; enter-set adds a set equal to another one - if both are set names,
; set-equal is called - if one is a quoted expression set, add-set is called

(defun enter-set (net a1 a2)
  (unless (immed-set a1) (change-property a1 'sort 'epilog::set))
  (unless (immed-set a2) (change-property a2 'sort 'epilog::set))
  (if (not (immed-set a1))
      (if (not (immed-set a2))
	  (sets-equal net a1 a2 t)
	  (add-set net a1 (get-set-immed a2))
	  )
      (if (not (immed-set a2))
	  (add-set net a2 (get-set-immed a1))
	  ; both set descriptions - cannot assert equal except in certain circ.
	  (check-equal-immed net a1 a2)
	  )
      )
  )

(defun check-equal-immed (net a1 a2 &aux
			      (a1mem (get-set-immed a1))
			      (a2mem (get-set-immed a2)))
  ; if either has just one member, then it is setting that member equal to
  ; all the members in the other set
  (check-equal-members net a1mem a2mem)
  )

;sets-subset takes current members of a1 and adds them to a2. a2 gets unfinished
; flag

(defun sets-subset (net s1 s2 &aux
			(set1 (getset net s1))
			set2 ts conn
			(mems1 (episet-contents set1)))
  (add-members net mems1 s2)
  (setq set2 (getset net s2))
  (unless (member-if #'(lambda (a) 
			 (and (equal (episet-connect-how a) 'subset-of)
			      (setq ts (findset net (car (episet-connect-sets a))))
			      (equal ts set2)))
		     (episet-connections set1))
	  (setq conn (make-episet-connect :main-set s1 :how 'subset-of 
				       :sets (list s2)))
	  (setq set1 (writeable-copy net set1))
	  (setq set2 (writeable-copy net set2))
	  (change-field 'episet-connections set1
			(append (episet-connections set1) (list conn)))
	  (change-field 'episet-connections-from set2
			(append (episet-connections-from set2) (list conn)))
	  (if (and (episet-member-type set2) (not (episet-member-type set1)))
	      (change-field 'episet-member-type set1 (episet-member-type set2)))
	  )
  )

(defun add-members (net mems s1 &aux
			(set1 (writeable-copy net (getset net s1)))
			(mems1 (episet-contents set1)))
  (setq mems1 (calc-union (list mems mems1)))
  (update-card net set1 mems1 t)
  )

; add-subset adds members from set a1 to set a2 because of subset-of relation

(defun add-subset (net a1 a2)
  (if (not (immed-set a1))
      (if (not (immed-set a2))
	  (sets-subset net a1 a2)
	  nil ; cannot assert s1 subset of (set a b c d)
	  )
      (if (not (immed-set a2))
	  (add-members net (get-set-immed a1) a2)
	  nil ; both set descriptions - cannot assert subset
	  )
      )
  )

(defun get-max-card (card &aux c)
  (if (numberp card)
      card
      (let ()
	   (add-interested-party card 'set-specialist)
	   (cond ((numberp (setq c (spec-eval-fn (make-fn-arg 'epilog::max-of card))))
		  c)
		 (t card))
	   )
      )
  )

(defun get-min-card (card &aux c)
  (if (numberp card)
      card
    (let ()
      (add-interested-party card 'set-specialist)
      (cond ((numberp (setq c (spec-eval-fn (make-fn-arg 'epilog::min-of card))))
	     c)
	    (t card))
      )
    )
  )

(defun add-cardinality (net setname card &aux
			    set1 oldmin oldmax min max nm)
  (when (not (immed-set setname))
	(setq set1 (writeable-copy net (getset net setname)))
	(setq oldmin (episet-cardinality-min set1))
	(setq oldmax (episet-cardinality-max set1))
	(setq min (get-min-card card))
	(setq max (get-max-card card))
	; if min and max both nil, ask # specialist for help
	; get min and max - add to interested party lists

	(if (or (and (numberp min) (numberp oldmax) (> min oldmax))
		(and (numberp max) (numberp oldmin) (< max oldmin))
		(and (numberp min) (numberp max) (< max min)))
	    nil
	(cond ((and (numberp min) 
		    (numberp max) 
		    (equal min max)) ; setting max and min
	    (when (and (or (not (equal oldmin oldmax)) (not (numberp oldmin)))
		       (or (not (numberp oldmin)) (>= min oldmin))
		       (or (not (numberp oldmax)) (<= max oldmax)))
		  (if (traced-p 'epilog::set-entry)
		      (print-line
			    "    Set Specialist: Setting cardinality of set "
				 setname " to " min))
		  (change-field 'episet-cardinality-min set1 min)
		  (change-field 'episet-cardinality-max set1 min)  ;card
		  ))
	      (t
	       (when (or (not (numberp oldmin)) 
			 (and (numberp min) (> min oldmin)))
		     (if (traced-p 'epilog::set-entry)
			 (print-line
		 "    Set Specialist: Setting cardinality lower bound of set "
		               setname " to " min))
	             (change-field 'episet-cardinality-min set1 min))
	
	       (when (or (not (numberp oldmax)) 
			 (and (numberp max) (< max oldmax)))
		     (if (traced-p 'epilog::set-entry)
			 (print-line
		 "    Set Specialist: Setting cardinality upper bound of set "
			       setname " to " max))
	             (change-field 'episet-cardinality-max set1 max)) )
	      ))
	(when (equal (episet-cardinality-max set1) 
		     (list-length (setq nm (remove 'epilog::+
						   (episet-contents set1)))))
	      (change-field 'episet-finished set1 t)
	      (change-field 'episet-contents set1 nm) )
	)
  )

(defun set-op-enter (net op pred arglist &aux
			 curset
			 (s1 (car arglist)))
  (unless (or (null op) (null pred) (null arglist) (cdr arglist))
	  ; ops coll, plur, plural, set-of-all
	  ; changes member-type of s1, also creates assertion
	  ;find s1; if does not exist, make it
	  (unless (immed-set s1)
	       (change-property s1 'sort 'set)
	       (setq curset (writeable-copy net (getset net s1)))
	       (change-field 'episet-member-type curset pred)
	       (when (traced-p 'epilog::set-entry)
		     (print-line  "    Set Specialist: Setting member type of set "
				  (print-info s1) " to " (print-info pred)))
	       )
	  (cond ((member op '(coll plur plural)) (add-type-assert s1 pred nil))
		((and (equal (op-type op) 'modified)
		      (equal (getp op 'operator) 'coll-of))
		 (add-type-assert s1 pred nil)
		 (coll-of-enter net op s1))
		)
	  )
  )

(defun coll-of-enter (net op s1 &aux
			  (opargs (getp op 'arguments))
			  (card (car opargs)))
  ; the coll-of operator acts on a number to give the cardinality of the set
  ; the predicate still defines the member type of the set - done already
  ; if opargs is a single number, the cardinality of s1 is exactly that number
  (when (and (null (cdr opargs))
	     (or (numberp card) (number-arg card)))
	(add-cardinality net s1 (car opargs)) )
  )

; enter enters the literal lit after first ensuring that it is in the
; set specialists domain and that it is not negated.

(defun enter (netname lit negp pred arglist &aux
		      (net (determine-net netname))
		      (a3 (caddr arglist))
		      )
  (declare (ignore lit))
  (when (and (not (non-constant-arg arglist))
	     (not negp)
	     (null a3))
	(enter-the-set-info net negp pred arglist))
  )

(defun enter-the-set-info (net negp pred arglist &aux
			       (reln (find-symbol (symbol-name pred) 'set))
			       (a1 (car arglist))
			       (a2 (cadr arglist))
			       (sort1 (find-sort a1))
			       (sort2 (find-sort a2))
			       op pr)
  (when (if (equal reln 'equal)
	    (or (equal sort1 'set) (equal sort2 'set))
	    t)   
	(when (traced-p 'epilog::set-entry)
	      (print-line "    Set Specialist: Entering " 
			  (if negp "NOT " "") (print-info a1) " " 
			(print-info pred) " " (if a2 (print-info a2) ""))
	      (if (not (equal (episet-subnet-identifier net) '/))   
		  (print-it "    into subnet " (episet-subnet-identifier net)))
	      )
	; if pred equal, only use for eval now, unless sets
	(cond ((and (symbolp pred)
		    (equal (getp pred 'pred-type) 'modified)
		    (setq op (getp pred 'operator))
		    (setq pr (car (getp pred 'arguments))))
	       (set-op-enter net op pr arglist)
	       )
	      ((equal reln 'equal)
	       (enter-equal net a1 a2))
	      ((and (equal reln 'has-cardinality)
		    (equal sort1 'epilog::set)
		    (member sort2 '(epilog::number epilog::integer)))
	       (add-cardinality net a1 a2))
	      ((and (equal reln 'member-of) (equal sort2 'epilog::set))
	       (add-member net a2 a1))
	      ((equal reln 'member-of-0)
	       ; ambiguous - cannot enter without making assumption
	       (print-line "    Set Specialist: member-of-0 is ambiguous - cannot enter ")
	       (print-line "       without making assumption - not entered") 
	       nil)
	      ((and (equal reln 'subset-of)
		    (equal sort1 'epilog::set)
		    (equal sort2 'epilog::set))
	       (add-subset net a1 a2))
	      )
	)
  )

; must have an empty set to use
(defvar e-s (getset (determine-net nil) 'epilog::empty-set))
(setf (episet-finished e-s) t)
(setf (episet-cardinality-min e-s) 0)
(setf (episet-cardinality-max e-s) 0)
